#  _CCARegressionTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from wai.test.decorators import RegressionTest

from wai.ma.algorithm.pls import CCARegression, DeflationMode
from wai.ma.core.matrix import Matrix

from ._AbstractPLSTest import AbstractPLSTest


class CCARegressionTest(AbstractPLSTest):
    """
    Test case for the NIPALS algorithm.
    """
    @classmethod
    def subject_type(cls):
        return CCARegression

    @RegressionTest
    def deflation_mode_canonical(self, subject: CCARegression, *resources: Matrix):
        subject.deflation_mode = DeflationMode.CANONICAL
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def deflation_mode_regression(self, subject: CCARegression, *resources: Matrix):
        subject.deflation_mode = DeflationMode.REGRESSION
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def norm_Y_weights_true(self, subject: CCARegression, *resources: Matrix):
        subject.norm_Y_weights = True
        return self.standard_regression(subject, *resources)



#  _AbstractTransformation.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from abc import abstractmethod, ABC

from ..core import Filter
from ..core.matrix import Matrix


class AbstractTransformation(Filter, ABC):
    def __init__(self):
        self.configured: bool = False

    def reset(self):
        self.configured = False

    @abstractmethod
    def configure(self, data: Matrix):
        pass

    @abstractmethod
    def do_transform(self, data: Matrix) -> Matrix:
        pass

    def transform(self, data: Matrix) -> Matrix:
        if not self.configured:
            self.configure(data)
        return self.do_transform(data)

    @abstractmethod
    def do_inverse_transform(self, data: Matrix) -> Matrix:
        pass

    def inverse_transform(self, data: Matrix) -> Matrix:
        if not self.configured:
            self.configure(data)
        return self.do_inverse_transform(data)

    @classmethod
    def quick_apply(cls, data: Matrix) -> Matrix:
        return cls().transform(data)




#  _Logger.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys


class Logger:
    """
    Class implementing equivalent functionality to java's
    java.util.logging.Logger class.
    """
    def __init__(self, name: str):
        self.__name = name

    @staticmethod
    def get_logger(name: str) -> 'Logger':
        # TODO
        return Logger(name)

    @staticmethod
    def warning(message: str):
        print(message, file=sys.stderr)

    @staticmethod
    def info(message: str):
        print(message, file=sys.stdout)



#  _PassThroughTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from wai.test.decorators import Test

from wai.ma.core.matrix import Matrix
from wai.ma.transformation import PassThrough

from ._AbstractTransformationTest import AbstractTransformationTest


class PassThroughTest(AbstractTransformationTest):
    @classmethod
    def subject_type(cls):
        return PassThrough

    @Test
    def result_unchanged(self, subject: PassThrough, bolts: Matrix):
        transform: Matrix = subject.transform(bolts)

        is_equal: bool = bolts.sub(transform).abs().all(lambda v: v < 1e-7)
        self.assertTrue(is_equal)



#  _QuantileTransformer.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from sklearn.preprocessing import QuantileTransformer as Intern

from ..core.matrix import Matrix
from ._AbstractTransformation import AbstractTransformation


class QuantileTransformer(AbstractTransformation):
    # The output distribution options
    OUTPUT_DISTRIBUTIONS = ("uniform", "normal")

    def __init__(self):
        super().__init__()

        self.n_quantiles: int = 1000
        self.output_distribution: str = QuantileTransformer.OUTPUT_DISTRIBUTIONS[0]
        self.ignore_implicit_zeroes: bool = False
        self.subsample: int = 100000
        self.random_state: Optional[int] = None

        self._intern: Optional[Intern] = None

    def __setattr__(self, key, value):
        # Validate values
        validator_name = 'validate_' + key
        if hasattr(self, validator_name):
            validator = getattr(self, validator_name)
            validator(value)

            # If validation passes, reset
            self.reset()

        super().__setattr__(key, value)

    def validate_output_distribution(self, selection: str):
        if selection not in QuantileTransformer.OUTPUT_DISTRIBUTIONS:
            raise ValueError(f"Output distribution was {selection} but must be one of: " +
                             ", ".join(QuantileTransformer.OUTPUT_DISTRIBUTIONS))

    def reset(self):
        super().reset()
        self._intern = None

    def configure(self, data: Matrix):
        self._intern = Intern(n_quantiles=self.n_quantiles,
                              output_distribution=self.output_distribution,
                              ignore_implicit_zeros=self.ignore_implicit_zeroes,
                              subsample=self.subsample,
                              random_state=self.random_state)
        self._intern.fit(data.data)
        self.configured = True

    def do_transform(self, data: Matrix) -> Matrix:
        return Matrix(self._intern.transform(data.data))

    def do_inverse_transform(self, data: Matrix) -> Matrix:
        return Matrix(self._intern.inverse_transform(data.data))



#  __init__.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Aliases
from ._AbstractTransformation import AbstractTransformation
from ._Center import Center
from ._FFT import FFT, OutputMode
from ._Log import Log
from ._PowerTransformer import PowerTransformer
from ._QuantileTransformer import QuantileTransformer
from ._RobustScaler import RobustScaler
from ._RowNorm import RowNorm
from ._Standardize import Standardize
from ._SavitzkyGolay import SavitzkyGolay
from ._SavitzkyGolay2 import SavitzkyGolay2
from ._MultiplicativeScatterCorrection import MultiplicativeScatterCorrection
from ._PassThrough import PassThrough

center = Center.quick_apply
row_norm = RowNorm.quick_apply
standardize = Standardize.quick_apply
log = Log.quick_apply
